"""Control worker: consumes from the Kafka ``control-events`` topic and
executes async side effects (S3 blob cleanup, cascade operations, etc.).
"""

from __future__ import annotations

import asyncio
from collections.abc import Callable
from typing import TYPE_CHECKING

from aiokafka import AIOKafkaConsumer
from loguru import logger
from matyan_api_models.kafka import ControlEvent
from pydantic import ValidationError

from matyan_backend.config import SETTINGS
from matyan_backend.fdb_types import FDBError
from matyan_backend.kafka.security import kafka_security_kwargs
from matyan_backend.storage import blob
from matyan_backend.storage.fdb_client import ensure_directories, init_fdb
from matyan_backend.workers.metrics import (
    CONTROL_EVENTS_CONSUMED,
    CONTROL_EVENTS_PROCESSED,
    CONTROL_PROCESSING_DURATION,
    CONTROL_PROCESSING_ERRORS,
)

if TYPE_CHECKING:
    from matyan_backend.fdb_types import Database


class ControlWorker:
    def __init__(self) -> None:
        self._consumer: AIOKafkaConsumer | None = None
        self._db: Database | None = None
        self._running = False

    async def start(self) -> None:
        self._db = init_fdb()
        ensure_directories(self._db)
        logger.info("FDB initialized")

        security_kwargs = kafka_security_kwargs()
        self._consumer = AIOKafkaConsumer(
            SETTINGS.kafka_control_events_topic,
            bootstrap_servers=SETTINGS.kafka_bootstrap_servers,
            group_id="control-workers",
            auto_offset_reset="earliest",
            enable_auto_commit=False,
            value_deserializer=lambda v: v.decode(),
            **security_kwargs,
        )
        assert self._consumer is not None
        await self._consumer.start()
        logger.info(
            "Control worker started (topic={}, group=control-workers)",
            SETTINGS.kafka_control_events_topic,
        )

        self._running = True
        try:
            await self._consume_loop()
        finally:
            await self.stop()

    async def stop(self) -> None:
        self._running = False
        if self._consumer is not None:
            await self._consumer.stop()
            logger.info("Control worker stopped")

    async def _consume_loop(self) -> None:
        assert self._consumer is not None
        async for record in self._consumer:
            CONTROL_EVENTS_CONSUMED.inc()
            try:
                event = ControlEvent.model_validate_json(record.value)
                with CONTROL_PROCESSING_DURATION.time():
                    await asyncio.to_thread(self._handle_event, event)
                await self._consumer.commit()
                CONTROL_EVENTS_PROCESSED.labels(event_type=event.type).inc()
            except ValidationError:
                logger.warning("Invalid control event, skipping", offset=record.offset, partition=record.partition)
                CONTROL_PROCESSING_ERRORS.labels(error_type="validation").inc()
            except FDBError:
                logger.exception("FDB error processing control event", offset=record.offset, partition=record.partition)
                CONTROL_PROCESSING_ERRORS.labels(error_type="fdb").inc()
            except Exception as e:  # noqa: BLE001
                err_type = type(e).__name__
                if err_type in ("ClientError", "GoogleCloudError", "HttpResponseError", "ResourceNotFoundError"):
                    logger.exception(
                        "Blob storage error processing control event",
                        offset=record.offset,
                        partition=record.partition,
                    )
                    CONTROL_PROCESSING_ERRORS.labels(error_type="blob_storage").inc()
                else:
                    logger.exception(
                        "Unexpected error processing control event",
                        offset=record.offset,
                        partition=record.partition,
                    )
                    CONTROL_PROCESSING_ERRORS.labels(error_type="unknown").inc()

    def _handle_event(self, event: ControlEvent) -> None:
        db = self._db
        assert db is not None
        handler = _HANDLERS.get(event.type)
        if handler is None:
            logger.warning("Unknown control event type: {}", event.type)
            return
        handler(db, event)


# ------------------------------------------------------------------
# Per-type handlers (all run synchronously inside asyncio.to_thread)
# ------------------------------------------------------------------





def _handle_run_deleted(db: Database, event: ControlEvent) -> None:  # noqa: ARG001
    p = event.payload
    run_id: str = p.get("run_id", "")
    if not run_id:
        logger.warning("run_deleted event missing run_id")
        return

    blob_keys: list[str] = p.get("blob_keys", [])

    deleted_count = blob.delete_blobs(blob_keys) if blob_keys else blob.delete_blob_prefix(f"{run_id}/")

    logger.info("run_deleted: cleaned up {} blob(s) for run {}", deleted_count, run_id)


def _handle_experiment_deleted(db: Database, event: ControlEvent) -> None:  # noqa: ARG001
    logger.info(
        "experiment_deleted: no-op (hook for future side effects), payload={}",
        event.payload,
    )


def _handle_tag_deleted(db: Database, event: ControlEvent) -> None:  # noqa: ARG001
    logger.info(
        "tag_deleted: no-op (hook for future side effects), payload={}",
        event.payload,
    )


def _handle_run_archived(db: Database, event: ControlEvent) -> None:  # noqa: ARG001
    logger.info(
        "run_archived: no-op (hook for future side effects), payload={}",
        event.payload,
    )


_HandlerFn = Callable[["Database", ControlEvent], None]

_HANDLERS: dict[str, _HandlerFn] = {
    "run_deleted": _handle_run_deleted,
    "experiment_deleted": _handle_experiment_deleted,
    "tag_deleted": _handle_tag_deleted,
    "run_archived": _handle_run_archived,
    "run_unarchived": _handle_run_archived,
}



"""Tests for workers/control.py handler functions.

S3 is mocked; FDB is not used by the handlers directly (they only do S3 cleanup).
"""

from __future__ import annotations

from datetime import UTC, datetime
from unittest.mock import MagicMock, patch

from matyan_api_models.kafka import ControlEvent

from matyan_backend.workers.control import (
    _HANDLERS,
    ControlWorker,
    _handle_experiment_deleted,
    _handle_run_archived,
    _handle_run_deleted,
    _handle_tag_deleted,
)

# ruff: noqa: SLF001


def _event(event_type: str, **payload: object) -> ControlEvent:
    return ControlEvent(
        type=event_type,
        timestamp=datetime.now(UTC),
        payload=dict(payload),
    )


class TestHandleRunDeleted:
    @patch("matyan_backend.workers.control.blob.delete_blobs")
    def test_with_explicit_blob_keys(self, mock_delete_blobs: MagicMock) -> None:
        db = MagicMock()
        _handle_run_deleted(db, _event("run_deleted", run_id="r1", blob_keys=["r1/a.png", "r1/b.wav"]))
        mock_delete_blobs.assert_called_once_with(["r1/a.png", "r1/b.wav"])

    @patch("matyan_backend.workers.control.blob.delete_blob_prefix")
    def test_with_prefix_listing(self, mock_delete_blob_prefix: MagicMock) -> None:
        db = MagicMock()
        _handle_run_deleted(db, _event("run_deleted", run_id="r2"))
        mock_delete_blob_prefix.assert_called_once_with("r2/")

    @patch("matyan_backend.workers.control.blob.delete_blob_prefix")
    @patch("matyan_backend.workers.control.blob.delete_blobs")
    def test_missing_run_id(self, mock_delete_blobs: MagicMock, mock_delete_blob_prefix: MagicMock) -> None:
        db = MagicMock()
        _handle_run_deleted(db, _event("run_deleted"))
        mock_delete_blobs.assert_not_called()
        mock_delete_blob_prefix.assert_not_called()


class TestHandleNoOpEvents:
    def test_experiment_deleted(self) -> None:
        _handle_experiment_deleted(MagicMock(), _event("experiment_deleted", experiment_id="e1"))

    def test_tag_deleted(self) -> None:
        _handle_tag_deleted(MagicMock(), _event("tag_deleted", tag_id="t1"))

    def test_run_archived(self) -> None:
        _handle_run_archived(MagicMock(), _event("run_archived", run_id="r1"))


class TestControlWorkerDispatch:
    def test_unknown_event_type(self) -> None:
        worker = ControlWorker()
        worker._db = MagicMock()
        worker._handle_event(_event("completely_unknown"))

    def test_dispatches_run_deleted(self) -> None:
        mock_handler = MagicMock()
        original = _HANDLERS["run_deleted"]
        _HANDLERS["run_deleted"] = mock_handler
        try:
            worker = ControlWorker()
            worker._db = MagicMock()
            event = _event("run_deleted", run_id="dispatch1")
            worker._handle_event(event)
            mock_handler.assert_called_once_with(worker._db, event)
        finally:
            _HANDLERS["run_deleted"] = original



"""Unified blob abstraction for S3, GCS, and Azure."""

from __future__ import annotations

from typing import TYPE_CHECKING

from matyan_backend.config import SETTINGS
from matyan_backend.storage import azure_client, gcs_client, s3_client

if TYPE_CHECKING:
    from collections.abc import Iterator


def get_blob(blob_key: str) -> bytes:
    """Fetch an object's body from the active blob backend as raw bytes."""
    match SETTINGS.blob_backend_type:
        case "azure":
            return azure_client.get_blob(blob_key)
        case "gcs":
            return gcs_client.get_blob(blob_key)
        case _:
            return s3_client.get_blob(blob_key)


def stream_blob(blob_key: str, chunk_size: int = 1024 * 1024) -> Iterator[bytes]:
    """Yield an object in fixed-size chunks from the active blob backend."""
    match SETTINGS.blob_backend_type:
        case "azure":
            yield from azure_client.stream_blob(blob_key, chunk_size)
        case "gcs":
            yield from gcs_client.stream_blob(blob_key, chunk_size)
        case _:
            yield from s3_client.stream_blob(blob_key, chunk_size)


def get_blob_size(blob_key: str) -> int:
    """Return the size in bytes of an object from the active blob backend."""
    match SETTINGS.blob_backend_type:
        case "azure":
            return azure_client.get_blob_size(blob_key)
        case "gcs":
            return gcs_client.get_blob_size(blob_key)
        case _:
            return s3_client.get_blob_size(blob_key)


def delete_blobs(keys: list[str]) -> int:
    """Delete a list of objects from the active blob backend. Returns count of deleted objects."""
    match SETTINGS.blob_backend_type:
        case "azure":
            return azure_client.delete_blobs(keys)
        case "gcs":
            return gcs_client.delete_blobs(keys)
        case _:
            return s3_client.delete_blobs(keys)


def delete_blob_prefix(prefix: str) -> int:
    """Delete all objects under a prefix from the active blob backend. Returns count of deleted objects."""
    match SETTINGS.blob_backend_type:
        case "azure":
            return azure_client.delete_blob_prefix(prefix)
        case "gcs":
            return gcs_client.delete_blob_prefix(prefix)
        case _:
            return s3_client.delete_blob_prefix(prefix)



"""Lazy S3 client for reading blob objects from S3-compatible storage."""

from __future__ import annotations

from typing import TYPE_CHECKING

import boto3
from botocore.config import Config as BotoConfig

from matyan_backend.config import SETTINGS

if TYPE_CHECKING:
    from collections.abc import Iterator

    from types_boto3_s3 import S3Client

_client: S3Client | None = None


def _get_client() -> S3Client:
    global _client  # noqa: PLW0603
    if _client is None:
        _client = boto3.client(
            "s3",
            endpoint_url=SETTINGS.s3_endpoint,
            aws_access_key_id=SETTINGS.s3_access_key,
            aws_secret_access_key=SETTINGS.s3_secret_key,
            config=BotoConfig(signature_version="s3v4"),
            region_name=SETTINGS.s3_region,
        )
    return _client


def get_blob(s3_key: str) -> bytes:
    """Fetch an object's body from S3 as raw bytes."""
    resp = _get_client().get_object(Bucket=SETTINGS.s3_bucket, Key=s3_key)
    return resp["Body"].read()


def stream_blob(s3_key: str, chunk_size: int = 1024 * 1024) -> Iterator[bytes]:
    """Yield an S3 object in fixed-size chunks (default 1 MB)."""
    resp = _get_client().get_object(Bucket=SETTINGS.s3_bucket, Key=s3_key)
    body = resp["Body"]
    while chunk := body.read(chunk_size):
        yield chunk


def get_blob_size(s3_key: str) -> int:
    """Return the size in bytes of an S3 object."""
    resp = _get_client().head_object(Bucket=SETTINGS.s3_bucket, Key=s3_key)
    return resp["ContentLength"]


def delete_blobs(keys: list[str]) -> int:
    """Delete multiple objects from S3 in batches. Returns count of deleted objects."""
    if not keys:
        return 0
    client = _get_client()
    bucket = SETTINGS.s3_bucket
    objects = [{"Key": k} for k in keys]
    deleted_count = 0
    # S3 maximum limits deletion to 1000 per request
    for i in range(0, len(objects), 1000):
        chunk = objects[i : i + 1000]
        resp = client.delete_objects(Bucket=bucket, Delete={"Objects": chunk, "Quiet": True})
        deleted_count += len(chunk) - len(resp.get("Errors", []))
    return deleted_count


def delete_blob_prefix(prefix: str) -> int:
    """Delete all objects under a prefix from S3. Returns count of deleted objects."""
    client = _get_client()
    bucket = SETTINGS.s3_bucket
    deleted_count = 0
    paginator = client.get_paginator("list_objects_v2")
    for page in paginator.paginate(Bucket=bucket, Prefix=prefix):
        contents = page.get("Contents", [])
        if not contents:
            continue
        objects = [{"Key": obj["Key"]} for obj in contents]
        client.delete_objects(Bucket=bucket, Delete={"Objects": objects, "Quiet": True})
        deleted_count += len(objects)
    return deleted_count



"""Lazy GCS client for reading blob objects from Google Cloud Storage."""

from __future__ import annotations

from typing import TYPE_CHECKING

from google.cloud import storage

from matyan_backend.config import SETTINGS

if TYPE_CHECKING:
    from collections.abc import Iterator

_client: storage.Client | None = None
_bucket: storage.Bucket | None = None


def _get_bucket() -> storage.Bucket:
    global _client, _bucket  # noqa: PLW0603
    if _client is None:
        _client = storage.Client()
        _bucket = _client.bucket(SETTINGS.gcs_bucket)
    return _bucket  # ty:ignore[invalid-return-type]


def get_blob(s3_key: str) -> bytes:
    """Fetch an object's body from GCS as raw bytes."""
    blob = _get_bucket().blob(s3_key)
    # Return empty bytes if not exists to mimic S3 client potential behavior, or let it raise.
    # The SDK raises google.cloud.exceptions.NotFound
    return blob.download_as_bytes()


def stream_blob(s3_key: str, chunk_size: int = 1024 * 1024) -> Iterator[bytes]:
    """Yield a GCS object in fixed-size chunks."""
    blob = _get_bucket().blob(s3_key)
    with blob.open("rb") as f:
        while chunk := f.read(chunk_size):
            yield chunk


def get_blob_size(s3_key: str) -> int:
    """Return the size in bytes of a GCS object."""
    blob = _get_bucket().get_blob(s3_key)
    if blob is None:
        return 0
    return blob.size or 0


def delete_blobs(keys: list[str]) -> int:
    """Delete multiple objects from GCS in batches. Returns count of deleted objects."""
    if not keys:
        return 0
    bucket = _get_bucket()
    deleted_count = 0
    # GCS batch delete
    for i in range(0, len(keys), 100):
        chunk = keys[i : i + 100]
        # delete_blobs accepts a list of blob names
        bucket.delete_blobs(chunk)
        deleted_count += len(chunk)
    return deleted_count


def delete_blob_prefix(prefix: str) -> int:
    """Delete all objects under a prefix from GCS. Returns count of deleted objects."""
    bucket = _get_bucket()
    blobs = list(bucket.list_blobs(prefix=prefix))
    if not blobs:
        return 0
    # Batch delete up to 100 at a time (or let SDK chunk it if supported,
    # but chunking explicitly is safer).
    deleted_count = 0
    for i in range(0, len(blobs), 100):
        chunk = blobs[i : i + 100]
        bucket.delete_blobs(chunk)
        deleted_count += len(chunk)
    return deleted_count



"""Download S3 artifacts for backed-up runs."""

from __future__ import annotations

from typing import TYPE_CHECKING

import boto3
from azure.identity import DefaultAzureCredential
from azure.storage.blob import BlobServiceClient
from botocore.config import Config as BotoConfig
from google.cloud import storage

from matyan_backend.config import SETTINGS

if TYPE_CHECKING:
    from pathlib import Path

    from types_boto3_s3 import S3Client


def _make_s3_client() -> S3Client:
    return boto3.client(
        "s3",
        endpoint_url=SETTINGS.s3_endpoint,
        aws_access_key_id=SETTINGS.s3_access_key,
        aws_secret_access_key=SETTINGS.s3_secret_key,
        config=BotoConfig(signature_version="s3v4"),
        region_name=SETTINGS.s3_region,
    )


def _make_gcs_client() -> storage.Client:
    return storage.Client()


def _make_azure_client() -> BlobServiceClient:
    if SETTINGS.azure_conn_str:
        return BlobServiceClient.from_connection_string(SETTINGS.azure_conn_str)
    return BlobServiceClient(
        account_url=SETTINGS.azure_account_url,
        credential=DefaultAzureCredential(),
    )


def export_blobs_for_run(  # noqa: C901
    run_hash: str,
    backup_dir: Path,
    *,
    s3_client: S3Client | None = None,
    gcs_client: storage.Client | None = None,
    azure_client: BlobServiceClient | None = None,
) -> tuple[int, int]:
    """Download all objects under ``<run_hash>/`` into the backup.

    Returns ``(blob_count, total_bytes)``.
    """
    prefix = f"{run_hash}/"
    blobs_dir = backup_dir / "runs" / run_hash / "blobs"

    blob_count = 0
    total_bytes = 0

    if SETTINGS.blob_backend_type == "gcs":
        client = gcs_client or _make_gcs_client()
        bucket = client.bucket(SETTINGS.gcs_bucket)
        for blob in bucket.list_blobs(prefix=prefix):
            blob_key = blob.name
            relative = blob_key[len(prefix) :]
            if not relative:
                continue

            local_path = blobs_dir / relative
            local_path.parent.mkdir(parents=True, exist_ok=True)

            data = blob.download_as_bytes()
            local_path.write_bytes(data)

            blob_count += 1
            total_bytes += len(data)

    elif SETTINGS.blob_backend_type == "azure":
        client = azure_client or _make_azure_client()
        container_client = client.get_container_client(SETTINGS.azure_container)
        for blob in container_client.list_blobs(name_starts_with=prefix):
            blob_key = blob.name
            relative = blob_key[len(prefix) :]
            if not relative:
                continue

            local_path = blobs_dir / relative
            local_path.parent.mkdir(parents=True, exist_ok=True)

            data = container_client.get_blob_client(blob.name).download_blob().readall()
            if isinstance(data, str):
                data = data.encode()
            local_path.write_bytes(data)

            blob_count += 1
            total_bytes += len(data)

    else:
        client = s3_client or _make_s3_client()
        bucket_name = SETTINGS.s3_bucket
        paginator = client.get_paginator("list_objects_v2")
        for page in paginator.paginate(Bucket=bucket_name, Prefix=prefix):
            for obj in page.get("Contents", []):
                blob_key = obj["Key"]
                relative = blob_key[len(prefix) :]
                if not relative:
                    continue

                local_path = blobs_dir / relative
                local_path.parent.mkdir(parents=True, exist_ok=True)

                resp = client.get_object(Bucket=bucket_name, Key=blob_key)
                data = resp["Body"].read()
                local_path.write_bytes(data)

                blob_count += 1
                total_bytes += len(data)

    return blob_count, total_bytes



"""Lazy Azure client for reading blob objects from Azure Blob Storage."""

from __future__ import annotations

from typing import TYPE_CHECKING, cast

from azure.identity import DefaultAzureCredential
from azure.storage.blob import BlobClient, BlobServiceClient
from loguru import logger

from matyan_backend.config import SETTINGS

if TYPE_CHECKING:
    from collections.abc import Iterator

_client: BlobServiceClient | None = None


def _get_client() -> BlobServiceClient:
    global _client  # noqa: PLW0603
    if _client is None:
        if SETTINGS.azure_conn_str:
            _client = BlobServiceClient.from_connection_string(SETTINGS.azure_conn_str)
        else:
            _client = BlobServiceClient(
                account_url=SETTINGS.azure_account_url,
                credential=DefaultAzureCredential(),
            )
    return _client


def _get_blob_client(blob_key: str) -> BlobClient:
    return _get_client().get_blob_client(container=SETTINGS.azure_container, blob=blob_key)


def get_blob(blob_key: str) -> bytes:
    """Fetch an object's body from Azure as raw bytes."""
    return cast("bytes", _get_blob_client(blob_key).download_blob().readall())


def stream_blob(blob_key: str, chunk_size: int = 1024 * 1024) -> Iterator[bytes]:  # noqa: ARG001
    """Yield an Azure object in chunks."""
    stream = _get_blob_client(blob_key).download_blob()
    yield from stream.chunks()


def get_blob_size(blob_key: str) -> int:
    """Return the size in bytes of an Azure object."""
    props = _get_blob_client(blob_key).get_blob_properties()
    return props.size


def delete_blobs(keys: list[str]) -> int:
    """Delete multiple objects from Azure. Returns count of deleted objects."""
    if not keys:
        return 0
    client = _get_client()
    container = client.get_container_client(SETTINGS.azure_container)
    count = 0
    for k in keys:
        try:
            container.delete_blob(k)
            count += 1
        except Exception:  # noqa: BLE001
            logger.warning("Failed to delete blob {} from Azure", k)
    return count


def delete_blob_prefix(prefix: str) -> int:
    """Delete all objects under a prefix from Azure. Returns count of deleted objects."""
    client = _get_client()
    container = client.get_container_client(SETTINGS.azure_container)
    blobs = list(container.list_blobs(name_starts_with=prefix))
    if not blobs:
        return 0
    count = 0
    for b in blobs:
        try:
            container.delete_blob(b.name)
            count += 1
        except Exception:  # noqa: BLE001
            logger.warning("Failed to delete blob {} from Azure", b.name)
    return count
